/*
 * Ultrasonic Anemometer Sketch, v0.1
 * Visit www.soldernerd.com for project details
 * Lukas Fässler
 * 2014-11-22
 */

#include <math.h>
#include <LiquidCrystal.h>

#define LCD_RS A1
#define LCD_RW A2
#define LCD_EN A3
#define LCD_D7 3
#define LCD_D6 4
#define LCD_D5 5
#define LCD_D4 6

LiquidCrystal lcd(LCD_RS, LCD_RW, LCD_EN, LCD_D4, LCD_D5, LCD_D6, LCD_D7);

#define AXIS 12
#define DIRECTION 11
#define SIGNAL 10
#define MUTE 9
#define ZEROCROSSING 8
#define TEMPERATURE_SENSOR A0
#define TMR1_COMPB_INT A5
#define TMR1_CAPTURE_INT A4
#define CALC 2

#define NORTH_TO_SOUTH 0 /* Direction=L, Axis=L */
#define EAST_TO_WEST 1 /* Direction=L, Axis=H */
#define SOUTH_TO_NORTH 2 /* Direction=H, Axis=L */
#define WEST_TO_EAST 3 /* Direction=H, Axis=H */

#define NUMBER_OF_ROUNDS 32
#define NUMBER_OF_ZEROCROSSINGS 16
#define DISTANCE 0.21

volatile uint8_t pulse_count;
volatile uint8_t axis_direction;
volatile uint8_t round_count;
volatile uint8_t measure_set;
volatile uint8_t calc_set;
volatile uint8_t done_flag;
volatile uint32_t envelope[2][4]; //32 bytes
volatile uint8_t zerocrossing_count;
volatile uint32_t zerocrossing[2][4][NUMBER_OF_ROUNDS]; //1024 bytes

int32_t envelope_tof_ns[4]; //16 bytes
int32_t zerocrossing_tof_ns[4]; //16 bytes

//Output Compare Register A for PWM: 399
#define OCR1AH_PWM 0x01
#define OCR1AL_PWM 0x8F
//Output Compare Register B for PWM: 199
#define OCR1BH_PWM 0x00
#define OCR1BL_PWM 0xC7
//Output Compare Register A long: 25199
#define OCR1AH_LONG 0x62
#define OCR1AL_LONG 0x6F
//Output Compare Register B long: 1999
#define OCR1BH_LONG 0x07
#define OCR1BL_LONG 0xCF


void setup()
{
  pulse_count = 0;
  axis_direction = 0;
  round_count = 0;
  measure_set = 0;
  calc_set = 1;
  done_flag = 0;
  pinMode(AXIS, OUTPUT); //Axis (NorthSouth vs EastWest)
  pinMode(DIRECTION, OUTPUT); // Direction (North->South vs South->North, East->West vs West->East)
  pinMode(SIGNAL, OUTPUT); //PWM Signal
  pinMode(MUTE, OUTPUT); // Amplifier Mute
  pinMode(TMR1_COMPB_INT, OUTPUT); 
  pinMode(TMR1_CAPTURE_INT, OUTPUT);
  pinMode(CALC, OUTPUT); 
  
  //Serial.begin(9600);
  
  /*
   * TCCR1A – Timer/Counter1 Control Register A, page 132
   * Bit 7:6 – COM1A1:0: Compare Output Mode for Channel A (Physical Pin 15, Arduino Pin 9)
   * Bit 5:4 – COM1B1:0: Compare Output Mode for Channel B (Physical Pin 16, Arduino Pin 10)
   * Bit 3:2 - Reserved
   * Bit 1:0 – WGM11:0: Waveform Generation Mode
   */
   
  TCCR1A = 0b00100011; //Output B, Fast PWM, Clear on Compare Match
  
  /*
   * TCCR1B – Timer/Counter1 Control Register B, page 134
   * Bit 7 – ICNC1: Input Capture Noise Canceler
   * Bit 6 – ICES1: Input Capture Edge Select
   * Bit 5 – Reserved
   * Bit 4:3 – WGM13:2: Waveform Generation Mode
   * Bit 2:0 – CS12:0: Clock Select
   */
  
  TCCR1B = 0b11011001; //Noise Canceler, Rising Edge, Prescaler=1, Fast PWM mode
  
  /*
   * TIMSK1 – Timer/Counter1 Interrupt Mask Register, page 136
   * Bit 7, 6 – Reserved
   * Bit 5 – ICIE1: Timer/Counter1, Input Capture Interrupt Enable
   * Bit 4, 3 – Reserved
   * Bit 2 – OCIE1B: Timer/Counter1, Output Compare B Match Interrupt Enable
   * Bit 1 – OCIE1A: Timer/Counter1, Output Compare A Match Interrupt Enable
   * Bit 0 – TOIE1: Timer/Counter1, Overflow Interrupt Enable
   */
   
  TIMSK1 = 0b00000100;
  
  /*
   * OCR1AH and OCR1AL – Output Compare Register 1 A, page 136
   * OCR1BH and OCR1BL – Output Compare Register 1 B, page 136
   */
  
  OCR1AH = OCR1AH_PWM;
  OCR1AL = OCR1AL_PWM;
  OCR1BH = OCR1BH_PWM;
  OCR1BL = OCR1BL_PWM;
  
  /*
   * ADCSRB – ADC Control and Status Register B, page 240
   * Bit 6 – ACME: Analog Comparator Multiplexer Enable
   */
   
  ADCSRB = 0b00000000;
  
  /*
   * ACSR – Analog Comparator Control and Status Register, page 240
   * Bit 7 – ACD: Analog Comparator Disable
   * Bit 6 – ACBG: Analog Comparator Bandgap Select
   * Bit 5 – ACO: Analog Comparator Output
   * Bit 4 – ACI: Analog Comparator Interrupt Flag
   * Bit 3 – ACIE: Analog Comparator Interrupt Enable
   * Bit 2 – ACIC: Analog Comparator Input Capture Enable
   * Bits 1, 0 – ACIS1, ACIS0: Analog Comparator Interrupt Mode Select
   */
   
  ACSR = 0b01000100; 
  
  lcd.begin(16, 2);
  lcd.setCursor(0, 0);
  lcd.print("Hey!");
  lcd.setCursor(0, 1);
  lcd.print("Whassup nerds?");
}


/*
 * Interrupt Service Routine for Timer 1 Output Compare Match B
 * Takes care of PWM Pulse generation, MUTE, AXIS, DIRECTION as well as some housekeeping
 */
ISR(TIMER1_COMPB_vect)
{
  digitalWrite(TMR1_COMPB_INT, HIGH);
  ++pulse_count;
  switch(pulse_count)
  {
    case(15):
    {
      // Pulses have been sent - Turn PWM output off
      TCCR1A = 0b00000011;
      // Set longer compare match and overflow values
      OCR1AH = OCR1AH_LONG;
      OCR1AL = OCR1AL_LONG;
      OCR1BH = OCR1BH_LONG;
      OCR1BL = OCR1BL_LONG;
      break;
    }
    case(16):
    {
      // Turn amplifier mute off - start listening, clear interrupt flag, enable envelope detector interrupt (via analog comparator)
      digitalWrite(MUTE, LOW);
      ACSR = 0b01000100; //Analog comparator as trigger source
      TIFR1 = 0b00100000; //Clear interrupt flag 
      TIMSK1 = 0b00100100; //Enable input capture interrupts
      zerocrossing_count = 0;
      // Reset output compare registers to PWM values
      OCR1AH = OCR1AH_PWM;
      OCR1AL = OCR1AL_PWM;
      OCR1BH = OCR1BH_PWM;
      OCR1BL = OCR1BL_PWM;
      break;
    }
    case(17):
    {
      ++axis_direction;
      axis_direction &= 0b11;
      // Will send another pulse soon, turn amplifier mute on 
      digitalWrite(MUTE, HIGH);
      // Set axis and direction
      digitalWrite(DIRECTION, (axis_direction&0b10));
      digitalWrite(AXIS, (axis_direction&0b01));
      // Increment round_count if all directions have been measured
      if(axis_direction==0b00)
      {
        ++round_count;
        //Measurement complete. Reset round_count, invert measure_set and calc_set, set done_flag;
        if(round_count==NUMBER_OF_ROUNDS)
        {
          round_count = 0;
          measure_set ^= 1;
          calc_set ^= 1;
          done_flag = 1;
          envelope[measure_set][NORTH_TO_SOUTH] = 192000; //15*400*32
          envelope[measure_set][EAST_TO_WEST] = 192000; //15*400*32
          envelope[measure_set][SOUTH_TO_NORTH] = 192000; //15*400*32
          envelope[measure_set][WEST_TO_EAST] = 192000; //15*400*32
        }
      }
      break;
    }
    case(18):
    {
      //Turn PWM output on and reset pulse_count
      TCCR1A = 0b00100011; 
      pulse_count = 0;
      break;
    }
  }
  digitalWrite(TMR1_COMPB_INT, LOW);
}

/*
 * Interrupt Service Routine for Timer 1 Input Capture
 * Takes care of actual measurements
 * Triggered by envelope detector (via analog comparator) and zero-crossing detector
 */
ISR(TIMER1_CAPT_vect)
{
  asm("sbi %0, %1 \n\t"::"I" (_SFR_IO_ADDR(PORTC)), "I" (PORTC4)); //Arduino Pin A4 high
  //digitalWrite(TMR1_CAPTURE_INT, HIGH);
  //Calculate timer value
  uint16_t timer_value;
  uint8_t timer_low;
  timer_low = ICR1L;
  timer_value = ICR1H;
  timer_value <<= 8;
  timer_value |= timer_low;
  //Envelope detector
  if(zerocrossing_count==0)
  {
      //Add measured counter value to envelope
      envelope[measure_set][axis_direction] += timer_value;
      //Reset zerocrossing
      zerocrossing[measure_set][axis_direction][round_count] = 72000; //96000-24000=72000
      //Set zero crossing detector as input capture trigger source
      ACSR = 0b01000000;  
      TCCR1B = 0b11011001; //Start with rising edge
  }
  //Zerocrossing detector
  else
  {
    //Add measured counter value to zerocrossing
    zerocrossing[measure_set][axis_direction][round_count] += timer_value;
    //Rising or falling edge?
    //Toggle bit 6 of the TCCR1B register (change between rising and falling edge capture)
    /*
    asm
    (
      "in __tmp_reg__, %0 \n\t"
      "ldi r16, 0x40 \n\t"
      "eor __tmp_reg__, r16 \n\t"
      "out %0, __tmp_reg__ \n\t"
      :
      :"I" (_SFR_IO_ADDR(TCCR1B))
    );
    */
    //Switch between rising and falling edge
    TCCR1B ^= 0b01000000;
    if(zerocrossing_count==NUMBER_OF_ZEROCROSSINGS)
    {
      //Disable Input Capture interrupts
      TIMSK1 = 0b00000100;
    }  
  }
  ++zerocrossing_count;
  //digitalWrite(TMR1_CAPTURE_INT, LOW);
  asm volatile("cbi %0, %1 \n\t"::"I" (_SFR_IO_ADDR(PORTC)), "I" (PORTC4)); //Arduino Pin A4 low
}

/*
 * Returns temperature in degrees centigrade as a float
 */
float temperature()
{
  // Reference Voltage = 5.0 Volts
  // Maximum DAC Value = 1023
  // OpAmp Gain = 4.3
  // Sensor Output = 0.01 Volts per Degree Centigrade
  // 5/(1023*4.3*0.01) = 0.11366478
  float temp = (float) analogRead(TEMPERATURE_SENSOR);
  return temp * 0.11366478;
}

double windspeed(int16_t timediff)
{
  if(timediff==0)
  {
     return 0.0; 
  }
  double temperature = 20.0;
  double sos = 331.5 + (0.6*temperature);
  double a = (double) timediff * -0.000000001;
  double b = 2 * DISTANCE;
  double c = a * pow(sos,2);
  double discriminant = (b*b)-(4*a*c);
  double wind = (-b+sqrt(discriminant)) / (2*a);
  return wind;
}

void loop()
{
  if(done_flag)
  {
    digitalWrite(CALC, HIGH);
    envelope_tof_ns[EAST_TO_WEST] = envelope[calc_set][EAST_TO_WEST];
    envelope_tof_ns[EAST_TO_WEST] *= 125;
    envelope_tof_ns[EAST_TO_WEST] >>= 6;
    envelope_tof_ns[SOUTH_TO_NORTH] = envelope[calc_set][SOUTH_TO_NORTH];
    envelope_tof_ns[SOUTH_TO_NORTH] *= 125;
    envelope_tof_ns[SOUTH_TO_NORTH] >>= 6;
    envelope_tof_ns[WEST_TO_EAST] = envelope[calc_set][WEST_TO_EAST];
    envelope_tof_ns[WEST_TO_EAST] *= 125;
    envelope_tof_ns[WEST_TO_EAST] >>= 6;
    
    for(uint8_t dir=0;dir<4;++dir)
    {
      //Calculate Envelope Detector Time-of-Flight (ToF) in Nanoseconds
      envelope_tof_ns[dir] = envelope[calc_set][dir];
      envelope_tof_ns[dir] *= 125;
      envelope_tof_ns[dir] >>= 6;
      
      //Calculate Zerocrossing Detector Time-of-Flight (ToF) in Nanoseconds
      zerocrossing_tof_ns[dir] = 0;
      uint32_t reference = envelope[calc_set][dir] >> 1;
      for(uint8_t i=0;i<NUMBER_OF_ROUNDS;++i)
      {
        while(zerocrossing[calc_set][dir][i] > reference)
          zerocrossing[calc_set][dir][i] -= 6400;
        zerocrossing_tof_ns[dir] += zerocrossing[calc_set][dir][i];
      }
      zerocrossing_tof_ns[dir] *= 125;
      zerocrossing_tof_ns[dir] >>= 10;  
    }
    
    //Calculate and display North-South wind speed
    lcd.setCursor(0, 0);
    lcd.print("N->S         m/s");
    int32_t diff_ns = zerocrossing_tof_ns[NORTH_TO_SOUTH]-zerocrossing_tof_ns[SOUTH_TO_NORTH];
    /*
    while(diff_ns<-12500)
      diff_ns+=25000;
    while(diff_ns>12500)
      diff_ns-=25000;
     */
    float wind_ns = windspeed(diff_ns);
    lcd.setCursor(5, 0);
    lcd.print(-wind_ns);
    
    //Calculate and display East-West wind speed
    lcd.setCursor(0, 1);
    lcd.print("E->W         m/s");
    int32_t diff_ew = zerocrossing_tof_ns[EAST_TO_WEST]-zerocrossing_tof_ns[WEST_TO_EAST];
    /*
    while(diff_ew<-12500)
      diff_ew+=25000;
    while(diff_ew>12500)
      diff_ew-=25000;
    */
    float wind_ew = windspeed(diff_ew);
    lcd.setCursor(5, 1);
    lcd.print(-wind_ew);
    
    done_flag = 0;
    digitalWrite(CALC, LOW);
  }
}

